"""课程相关的模块
"""
from __future__ import annotations
from typing import Any, Dict, Optional, Tuple, List, Union, ClassVar
# from pydantic.dataclasses import dataclass
import re
from datetime import date
from requests import Session, get
from ._lib_wrapper.dataclass import dataclass
from .utils.datetimes import parse_period_str, parse_weeks_str, parse_weekday_str, date_from_str
from .mycqu import MycquUnauthorized

__all__ = ("CQUSession", "CQUSessionInfo",
           "CourseTimetable", "CourseDayTime", "Course")


CQUSESSIONS_URL = "https://my.cqu.edu.cn/api/timetable/optionFinder/session?blankOption=false"
CUR_SESSION_URL = "https://my.cqu.edu.cn/api/resourceapi/session/cur-active-session"
TIMETABLE_URL = "https://my.cqu.edu.cn/api/timetable/class/timetable/student/table-detail"


@dataclass(order=True)
class CQUSession:
    """重大的某一学期
    """
    year: int
    """主要行课年份"""
    is_autumn: bool
    """是否为秋冬季学期"""
    SESSION_RE: ClassVar = re.compile("^([0-9]{4})年?(春|秋)$")
    CQUSESSION_MIN: ClassVar[CQUSession]
    """my.cqu.edu.cn 支持的最早学期"""

    def __post_init_post_parse__(self):
        if hasattr(CQUSession, "CQUSESSION_MIN"):
            if self < CQUSession.CQUSESSION_MIN:
                raise ValueError(
                    f"session should not be earlier than {CQUSession.CQUSESSION_MIN}")

    def get_id(self) -> int:
        """获取该学期在 my.cqu.edu.cn 中的 id

        >>> CQUSession(2021, True).get_id()
        1038

        :return: 学期的 id
        :rtype: int
        """
        return (self.year-1503)*2 + int(self.is_autumn) + 1

    @staticmethod
    def from_str(string: str) -> CQUSession:
        """从学期字符串中解析学期

        >>> CQUSession.from_str("2021春")
        CQUSession(year=2021, is_autumn=False)
        >>> CQUSession.from_str("2020年秋")
        CQUSession(year=2020, is_autumn=True)

        :param string: 学期字符串，如“2021春”、“2020年秋”
        :type string: str
        :raises ValueError: 字符串不是一个预期中的学期字符串时抛出
        :return: 对应的学期
        :rtype: CQUSession
        """
        match = CQUSession.SESSION_RE.match(string)
        if match:
            return CQUSession(
                year=match[1],
                is_autumn=match[2] == "秋"
            )
        else:
            raise ValueError(f"string {string} is not a session")

    @staticmethod
    def fetch() -> List[CQUSession]:
        """从 my.cqu.edu.cn 上获取各个学期

        :return: 各个学期组成的列表
        :rtype: List[CQUSession]
        """
        session_list = []
        for session in get(CQUSESSIONS_URL).json():
            try:
                session_list.append(CQUSession.from_str(session["name"]))
            except ValueError:
                pass
        return session_list


CQUSession.CQUSESSION_MIN = CQUSession(2020, True)


@dataclass
class CQUSessionInfo:
    """某学期的一些额外信息，目前只找到获取当个学期这些信息的 web api
    """
    session: CQUSession
    """对应的学期"""
    begin_date: date
    """学期的开始日期"""
    end_date: date
    """学期的结束日期"""

    @staticmethod
    def fetch(session: Session) -> CQUSessionInfo:
        """从 my.cqu.edu.cn 上获取当前学期的学期信息，需要登录并认证了 mycqu 的会话

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 认证
        :return: 本学期信息对象
        :rtype: CQUSessionInfo
        """
        resp = session.get(CUR_SESSION_URL)
        if resp.status_code == 401:
            raise MycquUnauthorized()
        data = resp.json()["data"]
        return CQUSessionInfo(
            session=CQUSession(year=data["year"],
                               is_autumn=data["term"] == "秋"),
            begin_date=date_from_str(data["beginDate"]),
            end_date=date_from_str(data["endDate"])
        )


@dataclass
class CourseDayTime:
    """课程一次的星期和节次
    """
    weekday: int
    """星期，0 为周一，6 为周日，此与 :attr:`datetime.date.day` 一致"""
    period: Tuple[int, int]
    """节次，第一个元素为开始节次，第二个元素为结束节次（该节次也包括在范围内）。
    只有一节课时，两个元素相同。
    """

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> Optional[CourseDayTime]:
        """从反序列化的（一个）课表 json 中获取课程的星期和节次

        :param data: 反序列化成字典的课表 json
        :type data: Dict[str, Any]
        :return: 若其中有课程的星期和节次则返回相应对象，否则返回 :obj:`None`
        :rtype: Optional[CourseDayTime]
        """
        if data.get("periodFormat") and data.get("weekDayFormat"):
            return CourseDayTime(
                weekday=parse_weekday_str(data["weekDayFormat"]),
                period=parse_period_str(data["periodFormat"])
            )
        return None


@dataclass
class Course:
    """与具体行课时间无关的课程信息
    """
    name: str
    """课程名称"""
    code: str
    """课程代码"""
    course_num: Optional[str]
    """教学班号，在无法获取时（如考表 :class:`.exam.Exam` 中）设为 :obj:`None`"""
    dept: str
    """开课学院"""
    credit: Optional[float]
    """学分，无法获取到则为 :obj:`None`（如在考表 :class:`.exam.Exam` 中）"""
    instructor: Optional[str]
    """教师"""
    session: Optional[CQUSession]
    """学期，无法获取时则为 :obj:`None`"""

    @ staticmethod
    def from_dict(data: Dict[str, Any],
                  session: Optional[Union[str, CQUSession]] = None) -> Course:
        """从反序列化的（一个）课表或考表 json 中返回课程

        :param data: 反序列化成字典的课表或考表 json
        :type data: Dict[str, Any]
        :param session: 学期字符串或学期对象，留空则尝试从 ``data`` 中获取
        :type session: Optional[Union[str, CQUSession]], optional
        :return: 对应的课程对象
        :rtype: Course
        """
        if session is None and not data.get("session") is None:
            session = CQUSession.from_str(data["session"])
        if isinstance(session, str):
            session = CQUSession.from_str(session)
        return Course(
            name=data["courseName"],
            code=data["courseCode"],
            course_num=data.get("classNbr"),
            dept=data.get(
                "courseDepartmentName") or data["courseDeptShortName"],
            credit=data.get("credit"),
            instructor=data.get("instructorName"),
            session=session,
        )


@dataclass
class CourseTimetable:
    """课表对象，一个对象存储有相同课程、相同行课节次和相同星期的一批行课安排
    """
    course: Course
    """对应的课程"""
    stu_num: int
    """学生数"""
    classroom: Optional[str]
    """行课地点，无则为 :obj:`None`"""
    weeks: List[Tuple[int, int]]
    """行课周数，列表中每个元组 (a,b) 代表一个周数范围 a~b（包含 a, b），在单独的一周则有 b=a"""
    day_time: Optional[CourseDayTime]
    """行课的星期和节次，若时间是整周（如真实地占用整周的军训和某些实习、虚拟地使用一周的思修实践）
    则为 :obj:`None`"""
    whole_week: bool
    """是否真实地占用整周（如军训和某些实习是真实地占用、思修实践是“虚拟地占用”）"""

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> CourseTimetable:
        """从反序列化的一个课表 json 中获取课表

        :param data: 反序列化成字典的课表 json
        :type data: Dict[str, Any]
        :return: 课表对象
        :rtype: CourseTimetable
        """
        return CourseTimetable(
            course=Course.from_dict(data),
            stu_num=data["selectedStuNum"],
            classroom=data["roomName"],
            weeks=parse_weeks_str(data.get("weeks")
                                  or data.get("teachingWeekFormat")),  # type: ignore
            day_time=CourseDayTime.from_dict(data),
            whole_week=bool(data["wholeWeekOccupy"])
        )

    @staticmethod
    def fetch(session: Session, id_: str, cqu_session: Optional[Union[CQUSession, str]] = None) -> List[CourseTimetable]:
        """从 my.cqu.edu.cn 上获取学生或老师的课表

        :param session: 登录了统一身份认证（:func:`.auth.login`）并在 mycqu 进行了认证（:func:`.mycqu.access_mycqu`）的 requests 会话
        :type session: Session
        :param id_: 学生或教师的学工号
        :type id_: str
        :param cqu_session: 需要获取课表的学期，留空获取当前年级的课表
        :type cqu_session: Optional[Union[CQUSession, str]], optional
        :raises MycquUnauthorized: 若会话未在 my.cqu.edu.cn 进行认证
        :return: 获取的课表对象的列表
        :rtype: List[CourseTimetable]
        """
        if cqu_session is None:
            cqu_session = CQUSessionInfo.fetch(session).session
        elif isinstance(cqu_session, str):
            cqu_session = CQUSession.from_str(cqu_session)
        resp = session.post(TIMETABLE_URL,
                            params={"sessionId": cqu_session.get_id()},
                            json=[id_],
                            )
        if resp.status_code == 401:
            raise MycquUnauthorized()
        return [CourseTimetable.from_dict(timetable) for timetable in resp.json()["classTimetableVOList"]
                if timetable["teachingWeekFormat"]
                ]



"""考试相关的模块
"""
from __future__ import annotations
from typing import Dict, Any, Optional, List
from datetime import date, time
import requests
from .course import Course
from .utils.datetimes import date_from_str, time_from_str
# from pydantic.dataclasses import dataclass
from ._lib_wrapper.dataclass import dataclass
from ._lib_wrapper.encrypt import pad, aes_ecb_encryptor

__all__ = ("Exam",)

__exam_encryptor = aes_ecb_encryptor("cquisse123456789".encode())
EXAM_LIST_URL = "https://my.cqu.edu.cn/api/exam/examTask/get-student-exam-list-outside"


def get_exam_raw(student_id: str) -> Dict[str, Any]:
    """获取考表的原始 json 数据（被反序列化为 python 字典对象）

    :param student_id: 学号
    :type student_id: str
    :return: 反序列化后的课表 json 数据
    :rtype: Dict[str, Any]
    """
    return requests.get(EXAM_LIST_URL,
                        params={"studentId":
                                __exam_encryptor(
                                    pad(student_id.encode())).hex().upper()
                                }
                        ).json()


@dataclass
class Invigilator:
    """监考员信息
    """
    name: str
    """监考员姓名"""
    dept: str
    """监考员所在学院（可能是简称，如 :obj:`"数统"`）"""

    @staticmethod
    def from_dict(data: Dict[str, Optional[str]]) -> Invigilator:
        """从反序列化后的 json 数据中一名正/副监考员的数据中生成 :class:`Invigilator` 对象。

        :param data: 反序列化后的 json 数据中的一次考试数据
        :type data: Dict[str, Optional[str]]
        :return: 对应的 :class:`Invigilator` 对象
        :rtype: Invigilator
        """
        return Invigilator(
            name=data["instructor"],  # type: ignore
            dept=data["instDeptShortName"]  # type: ignore
        )


@dataclass
class Exam:
    """考试信息
    """
    course: Course
    """考试对应的课程，其中学分 :attr:`credit`、教师 :attr:`instructor`、教学班号 :attr:`course_num` 可能无法获取（其值会设置为 :obj:`None`）"""
    batch: str
    """考试批次，如 :obj:`"非集中考试周"`"""
    batch_id: int
    """选课系统中考试批次的内部id"""
    building: str
    """考场楼栋"""
    floor: int
    """考场楼层"""
    room: str
    """考场地点"""
    stu_num: int
    """考场人数"""
    date: date
    """考试日期"""
    start_time: time
    """考试开始时间"""
    end_time: time
    """考试结束时间"""
    week: int
    """周次"""
    weekday: int
    """星期，0为周一，6为周日"""
    stu_id: str
    """考生学号"""
    seat_num: int
    """考生座号"""
    chief_invi: List[Invigilator]
    """监考员"""
    asst_invi: Optional[List[Invigilator]]
    """副监考员"""

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> Exam:
        """从反序列化后的 json 数据中的一次考试数据生成 :class:`Exam` 对象

        :param data: 反序列化后的 json 数据中的一次考试数据
        :type data: Dict[str, Any]
        :return: 对应的 :class:`Exam` 对象
        :rtype: [type]
        """
        course = Course.from_dict(data)
        return Exam(
            course=course,
            batch=data["batchName"],
            batch_id=data["batchId"],
            building=data["buildingName"],
            room=data["roomName"],
            floor=data["floorNum"],
            date=date_from_str(data["examDate"]),
            start_time=time_from_str(data["startTime"]),
            end_time=time_from_str(data["endTime"]),
            week=data["week"],
            weekday=int(data["weekDay"]) - 1,
            stu_id=data["studentId"],
            seat_num=data["seatNum"],
            stu_num=data["examStuNum"],
            chief_invi=[Invigilator.from_dict(invi)
                        for invi in data["simpleChiefinvigilatorVOS"]],
            asst_invi=data["simpleAssistantInviVOS"] and [Invigilator.from_dict(invi)
                                                          for invi in data["simpleAssistantInviVOS"]]
        )

    @staticmethod
    def fetch(student_id: str) -> List[Exam]:
        """从 my.cqu.edu.cn 上获取指定学生的考表

        :param student_id: 学生学号
        :type student_id: str
        :return: 本学期的考表
        :rtype: List[Exam]
        """
        return [Exam.from_dict(exam)
                for exam in get_exam_raw(student_id)["data"]["content"]]



"""统一身份认证相关的模块
"""
from typing import Dict, Optional, Callable, Union
import random
import re
from base64 import b64encode
from requests import Session, Response
from bs4 import BeautifulSoup  # type: ignore
from ._lib_wrapper.encrypt import pad, aes_cbc_encryptor

__all__ = ("NotAllowedService", "NeedCaptcha", "InvaildCaptcha",
           "IncorrectLoginCredentials", "UnknownAuthserverException", "NotLogined",
           "is_logined", "logout", "access_service", "login")

AUTHSERVER_URL = "http://authserver.cqu.edu.cn/authserver/login"
AUTHSERVER_CAPTCHA_DETERMINE_URL = "http://authserver.cqu.edu.cn/authserver/needCaptcha.html"
AUTHSERVER_CAPTCHA_IMAGE_URL = "http://authserver.cqu.edu.cn/authserver/captcha.html"
AUTHSERVER_LOGOUT_URL = "http://authserver.cqu.edu.cn/authserver/logout"
_CHAR_SET = 'ABCDEFGHJKMNPQRSTWXYZabcdefhijkmnprstwxyz2345678'


def _random_str(length: int):
    return ''.join(random.choices(_CHAR_SET, k=length))


_SALT_RE: re.Pattern = re.compile('var pwdDefaultEncryptSalt = "([^"]+)"')


class NotAllowedService(Exception):
    """试图认证不允许的服务时抛出
    """


class NeedCaptcha(Exception):
    """登录统一身份认证时需要输入验证码时拋出
    """

    def __init__(self, image: bytes, image_type: str, after_captcha: Callable[[str], Response]):
        super().__init__("captcha is needed")
        self.image: bytes = image
        """验证码图片文件数据"""
        self.image_type: str = image_type
        """验证码图片 MIME 类型"""
        self.after_captcha: Callable[[str], Response] = after_captcha
        """将验证码传入，调用以继续进行登陆"""


class InvaildCaptcha(Exception):
    """登录统一身份认证输入了无效验证码时抛出
    """

    def __init__(self):
        super().__init__("invaild captcha")


class IncorrectLoginCredentials(Exception):
    """使用无效的的登录凭据（如错误的用户、密码）
    """

    def __init__(self):
        super().__init__("incorrect username or password")


class UnknownAuthserverException(Exception):
    """登录或认证服务过程中未知错误
    """


class NotLogined(Exception):
    """未登陆或登陆过期的会话被用于进行需要统一身份认证登陆的操作
    """

    def __init__(self):
        super().__init__("not in logined status")

# from https://github.com/CQULHW/CQUQueryGrade


def get_formdata(html: str, username: str, password: str) -> Dict[str, Union[str, bytes]]:
    soup = BeautifulSoup(html, 'html.parser')

    errors = soup.find("div", {"id": "msg", "class": "errors"})
    if not errors is None:
        error_str = errors.text.strip()
        if error_str == "应用未注册\n不允许使用认证服务来认证您访问的目标应用。":
            raise NotAllowedService(error_str)
        raise UnknownAuthserverException(
            "Error message before login: "+errors)
    lt = soup.find("input", {"name": "lt"})['value']
    dllt = soup.find("input", {"name": "dllt"})['value']
    execution = soup.find("input", {"name": "execution"})['value']
    _event_id = soup.find("input", {"name": "_eventId"})['value']
    rm_shown = soup.find("input", {"name": "rmShown"})['value']
    salt_js = soup.find("script", {"type": "text/javascript"}).string
    assert (match := _SALT_RE.search(salt_js))
    key = match[1]  # 获取盐，被用来加密
    passwd_pkcs7 = pad((_random_str(64)+str(password)).encode())
    encryptor = aes_cbc_encryptor(key.encode(), _random_str(16).encode())
    passwd_encrypted = b64encode(encryptor(passwd_pkcs7))
    # 传入数据进行统一认证登录
    return {
        'username': username,
        'password': passwd_encrypted,
        'lt': lt,
        'dllt': dllt,
        'execution': execution,
        '_eventId': _event_id,
        'rmShown': rm_shown
    }


def is_logined(session: Session) -> bool:
    """判断是否处于统一身份认证登陆状态

    :param session: 会话
    :type session: Session
    :return: :obj:`True` 如果处于登陆状态，:obj:`False` 如果处于未登陆或登陆过期状态
    :rtype: bool
    """
    return session.get(AUTHSERVER_URL, allow_redirects=False).status_code == 302


def logout(session: Session) -> None:
    """注销统一身份认证登录状态

    :param session: 进行过登录的会话
    :type session: Session
    """
    session.get("http://authserver.cqu.edu.cn/authserver/logout")


def access_service(session: Session, service: str) -> Response:
    resp = session.get(AUTHSERVER_URL,
                       params={"service": service},
                       allow_redirects=False)
    if resp.status_code != 302:
        errors = BeautifulSoup(resp.text).find(
            "div", {"id": "msg", "class": "errors"})
        if not errors is None:
            raise NotAllowedService(errors.text[:-5].strip())
        raise NotLogined()
    return session.get(url=resp.headers['Location'], allow_redirects=False)


def login(session: Session,
          username: str,
          password: str,
          service: Optional[str] = None,
          timeout: int = 10,
          force_relogin: bool = False,
          captcha_callback: Optional[
              Callable[[bytes, str], Optional[str]]] = None
          ) -> Response:
    """登录统一身份认证

    :param session: 用于登录统一身份认证的会话
    :type session: Session
    :param username: 统一身份认证号或学工号
    :type username: str
    :param password: 统一身份认证密码
    :type password: str
    :param service: 需要登录的服务，默认（:obj:`None`）则先不登陆任何服务
    :type service: Optional[str], optional
    :param timeout: 连接超时时限，默认为 10（单位秒）
    :type timeout: int, optional
    :param force_relogin: 强制重登，当会话中已经有有效的登陆 cookies 时依然重新登录，默认为 :obj:`False`
    :type force_relogin: bool, optional
    :param captcha_callback: 需要输入验证码时调用的回调函数，默认为 :obj:`None` 即不设置回调；
                             当需要输入验证码，但回调没有设置或回调返回 :obj:`None` 时，抛出异常 :class:`NeedCaptcha`；
                             该函数接受一个 :class:`bytes` 型参数为验证码图片的文件数据，一个 :class:`str` 型参数为图片的 MIME 类型，
                             返回验证码文本或 :obj:`None`。
    :type captcha_callback: Optional[Callable[[bytes, str], Optional[str]]], optional
    :raises UnknownAuthserverException: 未知认证错误
    :raises InvaildCaptcha: 无效的验证码
    :raises IncorrectLoginCredentials: 错误的登陆凭据（如错误的密码、用户名）
    :raises NeedCaptcha: 需要提供验证码，获得验证码文本之后可调用所抛出异常的 :func:`NeedCaptcha.after_captcha` 函数来继续登陆
    :return: 登陆了统一身份认证后所跳转到的地址的 :class:`Response`
    :rtype: Response
    """
    login_page = session.get(
        url=AUTHSERVER_URL,
        params=None if service is None else {"service": service},
        allow_redirects=False,
        timeout=timeout)
    if login_page.status_code == 302:
        if not force_relogin:
            try:
                return login_page
            except NotLogined:
                pass
        else:
            logout(session)
    elif login_page.status_code != 200:
        raise UnknownAuthserverException()
    formdata = get_formdata(login_page.text, username, password)

    def after_captcha(captcha_str: Optional[str]):
        if captcha_str is None:
            if "captchaResponse" in formdata:
                del formdata["captchaResponse"]
        else:
            formdata["captchaResponse"] = captcha_str
        login_resp = session.post(
            url=AUTHSERVER_URL, data=formdata, allow_redirects=False)
        if login_resp.status_code != 302:
            soup = BeautifulSoup(login_resp.text, 'html.parser')
            errors = soup.find(
                "span", {"id": "msg", "class": "login_auth_error"})
            if errors is None:
                raise UnknownAuthserverException(
                    f"status code {login_resp.status_code} is got (302 expected) when sending login post, "
                    "but can not find the element span.login_auth_error#msg")
            else:
                error_str = errors.text.strip()
                if error_str == "无效的验证码":
                    raise InvaildCaptcha()
                elif error_str == "您提供的用户名或者密码有误":
                    raise IncorrectLoginCredentials()
                else:
                    raise UnknownAuthserverException(
                        f"status code {login_resp.status_code} is got (302 expected)"
                        f" when sending login post, {error_str}"
                    )
        return session.get(url=login_resp.headers['Location'], allow_redirects=False)

    captcha_str = None
    if session.get(AUTHSERVER_CAPTCHA_DETERMINE_URL, params={"username": username}).text == "true":
        captcha_img_resp = session.get(AUTHSERVER_CAPTCHA_IMAGE_URL)
        if captcha_callback is None:
            raise NeedCaptcha(captcha_img_resp.content,
                              captcha_img_resp.headers["Content-Type"],
                              after_captcha)
        captcha_str = captcha_callback(
            captcha_img_resp.content, captcha_img_resp.headers["Content-Type"])
        if captcha_str is None:
            raise NeedCaptcha(captcha_img_resp.content,
                              captcha_img_resp.headers["Content-Type"],
                              after_captcha)
    return after_captcha(captcha_str)



from typing import List, Tuple, Dict
from datetime import date, time
import pytz
TIMEZONE = pytz.timezone("Asia/Shanghai")
WEEKDAY: Dict[str, int] = {
    "一": 0,
    "二": 1,
    "三": 2,
    "四": 3,
    "五": 4,
    "六": 5,
    "日": 6
}


def time_from_str(string: str) -> time:
    hour, minute = map(int, string.split(":"))
    return time(hour, minute, second=0, tzinfo=TIMEZONE)


def parse_period_str(string: str) -> Tuple[int, int]:
    period = tuple(map(int, string.split("-")))
    assert len(period) == 1 or len(period) == 2
    return period[0], (period[1] if len(period) == 2 else period[0])


def parse_weeks_str(string: str) -> List[Tuple[int, int]]:
    return [parse_period_str(unit) for unit in string.split(',')]


def parse_weekday_str(string: str) -> int:
    return WEEKDAY[string]


def date_from_str(string: str) -> date:
    return date.fromisoformat(string)



"""my.cqu.edu.cn 认证相关的模块
"""
from typing import Dict
import re
from requests import Session
from .auth import access_service
__all__ = ("access_mycqu",)

MYCQU_TOKEN_INDEX_URL = "https://my.cqu.edu.cn/enroll/token-index"
MYCQU_TOKEN_URL = "https://my.cqu.edu.cn/authserver/oauth/token"
MYCQU_AUTHORIZE_URL = f"https://my.cqu.edu.cn/authserver/oauth/authorize?client_id=enroll-prod&response_type=code&scope=all&state=&redirect_uri={MYCQU_TOKEN_INDEX_URL}"
MYCQU_SERVICE_URL = "http://my.cqu.edu.cn/authserver/authentication/cas"
CODE_RE = re.compile(r"\?code=([^&]+)&")


class MycquUnauthorized(Exception):
    def __init__(self):
        super().__init__("Unanthorized in mycqu, auth.login firstly and then mycqu.access_mycqu")


def get_oauth_token(session: Session) -> str:
    # from https://github.com/CQULHW/CQUQueryGrade
    resp = session.get(MYCQU_AUTHORIZE_URL, allow_redirects=False)
    assert (match := CODE_RE.search(resp.headers['Location']))
    token_data = {
        'client_id': 'enroll-prod',
        'client_secret': 'app-a-1234',
        'code': match[1],
        'redirect_uri': MYCQU_TOKEN_INDEX_URL,
        'grant_type': 'authorization_code'
    }
    access_token = session.post(MYCQU_TOKEN_URL, data=token_data)
    return "Bearer " + access_token.json()['access_token']


def access_mycqu(session: Session, add_to_header: bool = True) -> Dict[str, str]:
    """用登陆了统一身份认证的会话在 my.cqu.edu.cn 进行认证

    :param session: 登陆了统一身份认证的会话
    :type session: Session
    :param add_to_header: 是否将 mycqu 的认证信息写入会话属性，默认为 :obj:`True`
    :type add_to_header: bool, optional
    :return: mycqu 认证信息的请求头，当 ``add_to_header`` 参数为 :obj:`True` 时无需手动使用该返回值
    :rtype: Dict[str, str]
    """
    session.get(access_service(session, MYCQU_SERVICE_URL).headers['Location'],
                allow_redirects=False)  # http 302 to https
    token = get_oauth_token(session)
    if add_to_header:
        session.headers["Authorization"] = token
    return {"Authorization": token}



"""从两种模块不同的模块名字中加载加密模块
"""
from typing import Callable
pad: Callable[[bytes], bytes]
aes_cbc_encryptor: Callable[[bytes, bytes], Callable[[bytes], bytes]]
aes_ecb_encryptor: Callable[[bytes], Callable[[bytes], bytes]]
try:
    from Cryptodome.Cipher import AES as AES_
    from Cryptodome.Util.Padding import pad as pad_
except (OSError, ImportError):
    try:
        from Crypto.Cipher import AES as AES__
        from Crypto.Util.Padding import pad as pad__
    except (OSError, ImportError):
        try:
            from pyaes.util import append_PKCS7_padding  # type: ignore
            from pyaes import AESModeOfOperationCBC  # type: ignore
            from pyaes import AESModeOfOperationECB

            def aes_cbc_encryptor(key, iv):
                encrypt = AESModeOfOperationCBC(key, iv).encrypt
                return lambda x: b''.join(encrypt(x[i: i+16]) for i in range(0, len(x), 16))

            def aes_ecb_encryptor(key):
                encrypt = AESModeOfOperationECB(key).encrypt
                return lambda x: b''.join(encrypt(x[i: i+16]) for i in range(0, len(x), 16))
            pad = append_PKCS7_padding
        except ImportError:
            raise ImportError(  # pylint: ignore disable=raise-missing-from
                "Please install pyryptodome, pyryptodomex or pyaes")
    else:
        def pad(x):
            return pad__(x, 16, style='pkcs7')

        def aes_cbc_encryptor(key, iv):
            return AES__.new(key=key, iv=iv, mode=AES__.MODE_CBC).encrypt

        def aes_ecb_encryptor(key):
            return AES__.new(key, AES__.MODE_ECB).encrypt
else:
    def pad(x):
        return pad_(x, 16, style='pkcs7')

    def aes_cbc_encryptor(key, iv):
        return AES_.new(key=key, iv=iv, mode=AES_.MODE_CBC).encrypt

    def aes_ecb_encryptor(key):
        return AES_.new(key, AES_.MODE_ECB).encrypt

__all__ = ("aes_cbc_encryptor", "aes_ecb_encryptor", "pad")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mycqu', 'mycqu._lib_wrapper', 'mycqu.utils']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4,<5', 'pydantic>=1,<2', 'pytz', 'requests>=2,<3']

setup_kwargs = {
    'name': 'mycqu',
    'version': '0.2.4.2',
    'description': '重庆重庆大学新教务网及相关 api 的封装',
    'long_description': '# pymycqu\n\n这个库对重庆大学 <https://my.cqu.edu.cn> 和统一身份认证的部分 web api 进行了封装，同时整理了相关数据模型。\n\nWork in progress... 欢迎反馈和补充\n\n感谢 <https://github.com/CQULHW/CQUQueryGrade> 项目提供了 <https://my.cqu.edu.cn> 的登陆方式。\n\n## 安装\n\n```bash\npip install mycqu\n```\n\n## 例子及文档\n\n见 <https://pymycqu.hagb.name>.\n\n## 许可\n\nAGPL 3.0\n',
    'author': 'Hagb',
    'author_email': 'hagb_green@qq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pymycqu.hagb.name',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""A workaround to make auto-completion work for dataclass of pydantic

https://github.com/samuelcolvin/pydantic/issues/650#issuecomment-709945440
"""

from typing import TYPE_CHECKING

# Trick
if TYPE_CHECKING:
    from dataclasses import dataclass
else:
    from pydantic.dataclasses import dataclass

__all__ = ("dataclass",)
